#!/usr/bin/env python

from __future__ import print_function

import sys
from struct import unpack
import logging
import gc
logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)

from .MPDULayer import MPDULayer
from .CPPDULayer import CPPDULayer
from .PPDULayer import PPDULayer
from .EMWINLayer import EMWINLayer
#from PrintLayer import PrintLayer
from .frametypes import VCDUMsg, MPDUMsg
from .opsmsgs import VCDUOpsMsg, MPDUOpsMsg


VCDU_SIZE = VCDUMsg.HEADER_SIZE_BYTES + VCDUMsg.DATA_SIZE_BYTES
VCDU_GC_INTERVAL = 10000
# protocol processing layers
mpduproc = MPDULayer()
# following are for virtual channels
cppduproc = [CPPDULayer() for n in range(63)]
ppduproc = PPDULayer()
emwinproc = EMWINLayer()
tot_vcdu = 0
missed_vcdu = 0

_last_tag = None

def report_error_counts():
    return (missed_vcdu, tot_vcdu)

def reset_error_counts():
    global missed_vcdu
    global tot_vcdu
    missed_vcdu = 0
    tot_vcdu = 0

# this function assumes tags have been un-pmt'd to [{offset: Class(key:KEY, value:VALUE)}]
def protoproc(data, tags, product_dir, product_dict):
    """Push data and tags through protocol processing"""
    global _last_tag
    global tot_vcdu
    global missed_vcdu

    # we assume frames are aligned with buffer
    # TODO: use nditer?
    for nframe, frame in enumerate(data[n:n + VCDU_SIZE] for n in range(0, data.size, VCDU_SIZE)):
        # either use new tag at the correct offset or _last_tag
        tag = tags.get(nframe * VCDU_SIZE, _last_tag)
        if tag and tag.key == 'ISYNC':
            frame = ~frame
        _last_tag = tag

        unlocked_vcdus = 0
        for offset, tag in tags.iteritems():
            if tag.key == 'lock':
                unlocked_vcdus+=tag.value/8192

         #process VCDU
        pdu = VCDUMsg()
        pdu.fromFrame(unpack('%dB' % len(frame), frame))
        pdu, msg = mpduproc.processData(pdu, VCDUOpsMsg())
        tot_vcdu += 1
        
        #Perform garbage collection every x vcdus.  Memory usage increases apparently unbounded otherwise.
        if tot_vcdu % VCDU_GC_INTERVAL == 0:
            gc.collect()

        #If more than 20 missed VCDUs, track misseed VCDUs from missed bytes
        if msg.missedFrames > 20:
            missed_vcdu += unlocked_vcdus
            tot_vcdu += unlocked_vcdus
        else:
            missed_vcdu += msg.missedFrames
            tot_vcdu += msg.missedFrames
        if not pdu:
            print("MPDU Error: VCID %d, %d missed, is %s frame" %
                  (msg.vcID, msg.missedFrames,
                  'new' if msg.isNewTransportFrame else 'not new'))
            continue
        #pdu, msg = cppduproc[msg.vcID].processData(pdu, msg)'
        c_pdu, c_msg, needs_data = cppduproc[msg.vcID].processData(pdu, msg)
        first_loop=True
        while not needs_data:
            empty_pdu = MPDUMsg()
            empty_pdu.fromFrame([0,0])
            empty_msg = MPDUOpsMsg()
            if not first_loop:
                c_pdu, c_msg, needs_data = cppduproc[msg.vcID].processData(empty_pdu, empty_msg)
            else:
                first_loop=False
            if not c_pdu:
                if c_msg.hasCrcErrors:
                    print('CRC errors')
                continue
            # returns LritFile object if complete, no_error = True if no error
            lritprod, no_error = ppduproc.processData(c_pdu, c_msg)
            if not lritprod:
                if not no_error:
                    print('Error decoding LRIT file')
                continue
            
            #check if product is in product filter
            ft = str(lritprod.filetype())
            if ft in product_dict.keys() and not product_dict[ft]:
                continue
            else:         
                if lritprod.filetype() == 214:
                    emwinprod,hasproduct = emwinproc.processData(lritprod,no_error)
                    if hasproduct:
                        emwinprod.toFile()
                else:
                    lritprod.toFile(product_dir)
            
def main(*args):
    """Create a protocol stack from the MPDU up and feed in from file"""

    with open(args[0], 'rb') as f:
        for data in iter(lambda: f.read(VCDU_SIZE), ""):
            if len(data) < VCDU_SIZE:
                print("Incomplete VCDU, skipping")
                continue

            vcdu = VCDUMsg()
            vcdu.fromFrame(unpack('%dB' % len(data), data))
            print("VCDU:")
            print(vcdu)
            mpdu, mpduops = mpduproc.processData(vcdu, VCDUOpsMsg())
            print("MPDU:")
            print(mpdu)
            cppdu, cppduops = cppduproc.processData(mpdu, mpduops)
            if not cppdu: continue
            print("CPPDU:")
            print(cppdu)
            ppdu, ppduops = ppduproc.processData(cppdu, cppduops)
            if not ppdu: continue
            print("PPDU:")
            print(ppdu)



if __name__ == '__main__':
    sys.exit(main(*sys.argv[1:]))
